package stundentafel_cnf;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Random;
import java.util.TreeSet;
import javax.swing.SwingUtilities;

/**
 * @author Benjamin A. Bartsch
 */
public final class D_Berechnung implements Comparable<D_Berechnung>, Runnable {

    final int beID;
    private final Data data2;

    Process process;
    private ByteArrayOutputStream baos;
    private final Random rnd = new Random();
    private final LinkedList<Integer> listRND = new LinkedList<>();
    private final LinkedList<String> commands = new LinkedList<>();
    private final HashMap<String, D_Fach_SAT> mapFa = new HashMap<>();

    int c_vars;
    int c_clauses;
    private final int c_TRUE;
    private final int c_FALSE;
    private String ergebnis = "???";
    private boolean[] state = null;

    LinkedList<String> html = new LinkedList<>();

    @SuppressWarnings("CallToThreadStartDuringObjectConstruction")
    D_Berechnung(int id, Data data, String header) throws IOException, UserException, InternalException, ParseException {
        // Checks
        if (Static.get_path_crypto().trim().isEmpty()) {
            throw new UserException("Sie müssen den Pfad des Solvers angeben!");
        }

        // Copy
        this.data2 = data;
        this.beID = id;
        this.commands.addLast(Static.get_path_crypto());
        this.commands.addLast("--threads=1");

        // Starte den Solver-Prozess nur dann, wenn der header bekannt ist.
        if (header != null) {
            ProcessBuilder pb = new ProcessBuilder(commands);
            pb.redirectErrorStream(true);
            process = pb.start();
            new Thread(this).start();
            baos = new ByteArrayOutputStream();
            write_bytes(header, true);
        }

        c_TRUE = c_new_var();
        c_1(c_TRUE);
        c_FALSE = -c_TRUE;

        this.cnf_create(data);
        this.cnf_pre(data);
        this.cnf_post(data);

        // Ausgabe
        html.addLast("<!DOCTYPE html>");
        html.addLast("<html>");
        html.addLast("<body>");

        html.addLast("<h1>Regeln</h1>");
        if (!data.co_get_regeln().isEmpty()) {
            String lines = Static.htmlEscapeSpitzeKlammern(data.co_get_regeln());
            lines = Static.htmlEscapeZeilenumbruchZuBr(lines);
            html.addLast("" + "<br>");
            html.addLast("" + lines);
        }

        // solver start signalisieren
        if (header != null) {
            write_bytes(null, true);
            process.getOutputStream().close();
        }
    }

    @Override
    public void run() {
        try {
            InputStream out = process.getInputStream();
            BufferedReader br = new BufferedReader(new InputStreamReader(out, "UTF-8"));
            while (process.isAlive() || br.ready()) {
                String line = br.readLine();
                if (line == null) {
                    break;
                }
                if (line.startsWith("v")) {
                    if (state == null) {
                        state = new boolean[c_vars + 1];
                    }
                    String[] split = line.split("\\s+");
                    for (int i = 1; i < split.length; i++) {
                        int k = Integer.parseInt(split[i].trim());
                        if (k > 0) {
                            state[k] = true;
                        }
                    }
                } else {
                    // System.out.println("SOLVER: " + line);
                }
            }
            process.waitFor();

            html.addLast("");

            // erzeuge ergebnis
            if (state == null) {
                html.addLast("Die derzeitigen Einstellungen sind unlösbar!");
            } else {
                erzeuge_gesamtausgabe();
            }

            ergebnis = "";
            for (String s : html) {
                ergebnis = ergebnis + s + "\n";
            }

            // informiere
            //data.beB_finished(beID);
        } catch (IOException | InterruptedException | NumberFormatException | ParseException ex) {
            ergebnis = "" + ex.getMessage();
            //data.beB_finished(beID);
            ex.printStackTrace();
        }

        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                data2.beB_finished(beID);
            }
        });

    }

    @Override
    public String toString() {
        return beID + ", V=" + c_vars + ", C=" + c_clauses;
    }

    @Override
    public int compareTo(D_Berechnung o) {
        return new Integer(beID).compareTo(o.beID);
    }

    private void cnf_create(Data data) throws UserException {

        for (int faID : data.fa_all().keySet()) {
            String faKuerzel = data.fa_single_get_kuerzel(faID);
            String faName = data.fa_single_get_name(faID);
            D_Fach_SAT fa = new D_Fach_SAT(faKuerzel, faName);
            mapFa.put(faKuerzel, fa);
        }

    }

    private void cnf_pre(Data data) throws IOException, InternalException, UserException, ParseException {
        // Fächer
        for (D_Fach_SAT fa : mapFa.values()) {
            fa.cnf_pre();
        }

        // Regeln
        for (String line : data.co_get_regel_liste()) {
            String[] split2 = line.split("\\#");
            int nr = Integer.parseInt(split2[0]);
            cnf_pre_regel(nr, split2[1]);
        }
    }

    private void cnf_post(Data data) throws IOException, InternalException, UserException, ParseException {
        // Fächer
        for (D_Fach_SAT fa : mapFa.values()) {
            fa.cnf_post();
        }

        // Regeln
        for (String line : data.co_get_regel_liste()) {
            String[] split2 = line.split("\\#");
            int nr = Integer.parseInt(split2[0]);
            cnf_post_regel(nr, split2[1]);
        }
    }

    // ########## REGELN - ANFANG ##########
    private void cnf_pre_regel(int nr, String text) throws UserException, IOException {
        // TODO
    }

    private void cnf_post_regel(int nr, String text) throws IOException, InternalException, UserException, ParseException {
        switch (nr) {
            case 100: // 100#Fach<A> hat in der Sek.<B> in der Summe<C> Unterrichtsstunden.
                cnf_post_regel_100_fach_a_in_der_sek_b_in_der_summe_c_unterrichtsstunden(Static.extract(text, 1), Static.extract(text, 2), Static.extract(text, 3));
                break;
            case 101: // 101#Fach<A> in der Stufe<B> ist <C>-stündig.
                cnf_post_regel_101_fach_a_in_der_stufe_b_ist_c_stuendig(Static.extract(text, 1), Static.extract(text, 2), Static.extract(text, 3));
                break;
            case 102: // 102#Fach<A> in der Sek.<B> niemals <C>-stündig.
                cnf_post_regel_102_fach_a_in_der_sek_b_niemals_c_stuendig(Static.extract(text, 1), Static.extract(text, 2), Static.extract(text, 3));
                break;
            case 103: // 103#Stufe<A> hat zwischen <B> und <C> Wochenstunden.
                cnf_post_regel_103_stufe_a_hat_zwischen_b_und_c_wochenstunden(Static.extract(text, 1), Static.extract(text, 2), Static.extract(text, 3));
                break;
            case 104: // 104#Stufe<A> gleichbehandlung der Fächer <B>, <C>, <D>.
                cnf_post_regel_104_stufe_a_gleichbehandlung_der_faecher_b_c_d(Static.extract(text, 1), Static.extract(text, 2), Static.extract(text, 3), Static.extract(text, 4));
                break;

            default:
                throw new UserException("Die Regel mit Nummer " + nr + " ist unbekannt!");
        }

    }

    private D_Fach_SAT cnf_post_get_fach(String a) throws UserException {
        if (!mapFa.containsKey(a)) {
            throw new UserException("Das Fach " + a + " existiert nicht!");
        }
        D_Fach_SAT fa = mapFa.get(a);
        return fa;
    }

    private void cnf_post_regel_100_fach_a_in_der_sek_b_in_der_summe_c_unterrichtsstunden(String a, String b, String c) throws IOException, InternalException, UserException {
        D_Fach_SAT fa = cnf_post_get_fach(a);
        fa.cnf_in_der_sek_b_in_der_summe_c_unterrichtsstunden(b, c);
    }

    private void cnf_post_regel_101_fach_a_in_der_stufe_b_ist_c_stuendig(String a, String b, String c) throws UserException, IOException, InternalException {
        D_Fach_SAT fa = cnf_post_get_fach(a);
        fa.cnf_in_der_stufe_b_ist_c_stuendig(b, c);
    }

    private void cnf_post_regel_102_fach_a_in_der_sek_b_niemals_c_stuendig(String a, String b, String c) throws UserException, IOException, InternalException {
        D_Fach_SAT fa = cnf_post_get_fach(a);
        fa.cnf_in_der_sek_b_niemals_c_stuendig(b, c);
    }

    private void cnf_post_regel_103_stufe_a_hat_zwischen_b_und_c_wochenstunden(String a, String b, String c) throws IOException, InternalException {
        int s = Integer.parseInt(a) - 5;
        LinkedList<D_Number_SAT> list = new LinkedList<>();
        for (D_Fach_SAT fa : mapFa.values()) {
            list.addLast(fa.stufe[s]);
        }

        while (list.size() >= 2) {
            D_Number_SAT n1 = list.pollFirst();
            D_Number_SAT n2 = list.pollFirst();
            D_Number_SAT n3 = n1.add(n2);
            list.addLast(n3);
        }
        list.pollFirst().cnf_zwischen_b_und_c_wochenstunden(b, c);
    }

    private void cnf_post_regel_104_stufe_a_gleichbehandlung_der_faecher_b_c_d(String a, String b, String c, String d) throws UserException, IOException, InternalException {
        int s = Integer.parseInt(a) - 5;
        D_Fach_SAT fa_b = cnf_post_get_fach(b);
        D_Fach_SAT fa_c = cnf_post_get_fach(c);
        fa_b.stufe[s].cnf_gleich(fa_c.stufe[s]);
        if (!d.equals(Static.WILDCARD)) {
            D_Fach_SAT fa_d = cnf_post_get_fach(d);
            fa_b.stufe[s].cnf_gleich(fa_d.stufe[s]);
        }
    }

    // ########## REGELN - ENDE ##########
    private void erzeuge_gesamtausgabe() throws ParseException, IOException {
        html.addLast("" + "<br>");

        String[] header = new String[]{"Fach", "5", "6", "7", "8", "9", "10", "Summe"};
        String[][] planS = new String[mapFa.size() + 1][8];

        int r = 0;
        for (D_Fach_SAT fa : new TreeSet<>(mapFa.values())) {
            planS[r][0] = fa.name;
            int summe = 0;
            for (int i = 0; i < 6; i++) {
                summe += fa.stufe[i].getNr();
                planS[r][i + 1] = fa.stufe[i].getNr() + "";
            }
            planS[r][7] = summe + "";
            r++;
        }

        int gesamtsumme = 0;
        for (int c = 0; c < planS[r].length; c++) {
            if (c == 0) {
                planS[r][c] = "Summe";
            }
            if ((c > 0) && (c <= 6)) {
                int summe = 0;
                for (D_Fach_SAT fa : new TreeSet<>(mapFa.values())) {
                    summe += fa.stufe[c - 1].getNr();
                }
                gesamtsumme += summe;
                planS[r][c] = "" + summe;
            }
        }
        planS[r][7] = "" + gesamtsumme;
        r++;

        Static.toHtmlTable(planS, header, html);

    }

    String get_ergebnis() {
        LinkedList<String> temp = new LinkedList<>();
        temp.addLast("<!DOCTYPE html>");
        temp.addLast("<html>");
        temp.addLast("<body>");
        temp.addAll(html);
        temp.addLast("</body>");
        temp.addLast("</html>");
        temp = Static.htmlEscape(temp);

        StringBuilder sb = new StringBuilder();
        for (String line : temp) {
            sb.append(line);
        }

        return sb.toString();

    }

    // ########################################################################
    int c_new_var() {
        if (listRND.isEmpty()) {
            for (int i = 0; i < 10000; i++) {
                c_vars++;
                if (rnd.nextBoolean()) {
                    listRND.addLast(+c_vars);
                } else {
                    listRND.addLast(-c_vars);
                }
            }
            Collections.shuffle(listRND, rnd);
        }
        return listRND.pollFirst();
    }

    int[] c_new_var_array(int size
    ) {
        int[] a = new int[size];
        for (int i = 0; i < a.length; i++) {
            a[i] = c_new_var();
        }
        return a;
    }

    int[][] c_new_var_matrix(int rows, int cols) {
        int[][] m = new int[rows][cols];
        for (int r = 0; r < rows; r++) {
            for (int c = 0; c < cols; c++) {
                m[r][c] = c_new_var();
            }
        }
        return m;
    }

    void write_bytes(String line, boolean send) throws IOException {
        if (baos != null) {
            if (line != null) {
                baos.write(line.getBytes());
            }
            if ((baos.size() > 1024 * 1024 * 32) || send) {
                process.getOutputStream().write(baos.toByteArray());
                baos.reset();
            }
        }
    }

    void c_1(int x) throws IOException, InternalException {
        String line = x + " 0" + "\n";
        if (x == 0) {
            throw new InternalException("Klausel " + line + " ist ungültig!");
        }
        c_clauses++;
        write_bytes(line, false);
    }

    void c_2(int x, int y) throws IOException, InternalException {
        String line = x + " " + y + " 0" + "\n";
        if ((x == 0) || (y == 0)) {
            throw new InternalException("Klausel " + line + " ist ungültig!");
        }
        c_clauses++;
        write_bytes(line, false);
    }

    void c_3(int x, int y, int z) throws IOException, InternalException {
        String line = x + " " + y + " " + z + " 0" + "\n";
        if ((x == 0) || (y == 0) || (z == 0)) {
            throw new InternalException("Klausel " + line + " ist ungültig!");
        }
        c_clauses++;
        write_bytes(line, false);
    }

    int c_at_most_one_tree(LinkedList<Integer> list) throws IOException, InternalException {
        while (list.size() >= 2) {
            int a = list.pollFirst();
            int b = list.pollFirst();
            c_2(-a, -b); // not both
            int c = c_new_var_OR(a, b);
            list.addLast(c);
        }
        if (list.isEmpty()) {
            return c_FALSE;
        }
        return list.pollFirst();
    }

    int c_new_var_OR(int x, int y) throws IOException, InternalException {
        int z = c_new_var();
        c_2(-x, z);
        c_2(-y, z);
        c_3(x, y, -z);
        return z;
    }

    void c_equal(int x, int y) throws IOException, InternalException {
        c_2(-x, +y);
        c_2(+x, -y);
    }

    void c_exactly_one_tree(int[] a) throws IOException, InternalException {
        LinkedList<Integer> list = new LinkedList<>();
        for (int i = 0; i < a.length; i++) {
            list.addLast(a[i]);
        }
        c_exactly_GENERIC(list, 1);
    }

    int c_exactly_one_tree(LinkedList<Integer> list) throws IOException, InternalException {
        int var = c_at_most_one_tree(list);
        c_1(var);
        return var;
    }

    int c_OR_tree(LinkedList<Integer> list) throws IOException, InternalException {
        while (list.size() >= 2) {
            int a = list.pollFirst();
            int b = list.pollFirst();
            int c = c_new_var_OR(a, b);
            list.addLast(c);
        }
        if (list.isEmpty()) {
            return c_FALSE;
        }
        return list.pollFirst();
    }

    int c_new_var_AND(int x, int y) throws IOException, InternalException {
        int c = c_new_var();
        c_2(x, -c);
        c_2(y, -c);
        c_3(-x, -y, c);
        return c;
    }

    boolean c_get(int var
    ) {
        if (var > 0) {
            return state[var];
        }
        return !state[-var];
    }

    void c_exactly_GENERIC(LinkedList<Integer> list, int amount) throws IOException, InternalException {
        list = new LinkedList<>(list);
        // special case: 0
        if (amount == 0) {
            for (int x : list) {
                c_1(-x);
            }
            return;
        }
        // special case: N
        if (amount == list.size()) {
            for (int x : list) {
                c_1(+x);
            }
            return;
        }
        // special case: 1
        if (amount == 1) {
            int var = c_exactly_one_tree(list);
            c_1(+var);
            return;
        }
        // else choose...
        c_exactly_NETWORK(list, amount);
    }

    void c_at_least_GENERIC(LinkedList<Integer> list, int minimum) throws IOException, InternalException {
        list = new LinkedList<>(list);
        // special case: trivial
        if (minimum <= 0) {
            return;
        }
        // special case: all true
        if (minimum == list.size()) {
            for (int x : list) {
                c_1(+x);
            }
            return;
        }
        // special case: at_least_one
        if (minimum == 1) {
            int var = c_OR_tree(list);
            c_1(+var);
            return;
        }
        // else choose...
        c_at_least_NETWORK(list, minimum);

    }

    void c_at_most_GENERIC(LinkedList<Integer> list, int maximum) throws IOException, InternalException {
        list = new LinkedList<>(list);
        // special case: trivial
        if (maximum >= list.size()) {
            return;
        }
        // special case: all false
        if (maximum == 0) {
            for (int x : list) {
                c_1(-x);
            }
            return;
        }
        // special case: at_most_one
        if (maximum == 1) {
            c_at_most_one_tree(list);
            return;
        }
        // else choose...
        c_at_most_NETWORK(list, maximum);

    }

    private void c_at_most_NETWORK(LinkedList<Integer> list, int maximum) throws IOException, InternalException {
        // sort
        c_bitonic_sort(list);
        // force FALSE
        for (int i = maximum; i < list.size(); i++) {
            c_1(-list.get(i));
        }
    }

    private void c_at_least_NETWORK(LinkedList<Integer> list, int minimum) throws IOException, InternalException {
        // sort
        c_bitonic_sort(list);
        // force FALSE
        for (int i = 0; i < minimum; i++) {
            c_1(+list.get(i));
        }
    }

    private void c_exactly_NETWORK(LinkedList<Integer> list, int amount) throws IOException, InternalException {
        // sort
        c_bitonic_sort(list);
        // force TRUE / FALSE
        for (int i = 0; i < list.size(); i++) {
            if (i < amount) {
                c_1(+list.get(i));
            } else {
                c_1(-list.get(i));
            }
        }
    }

    void c_bitonic_sort(LinkedList<Integer> list) throws IOException, InternalException {
        c_fill_False_until_power_two(list);
        c_bitonic_sort_power_two(list);
    }

    private void c_fill_False_until_power_two(LinkedList<Integer> list) throws IOException, InternalException {
        int size = 1;
        while (size < list.size()) {
            size *= 2;
        }
        while (list.size() < size) {
            list.addLast(c_FALSE);
        }
    }

    private void c_bitonic_sort_power_two(LinkedList<Integer> list) throws IOException, InternalException {
        for (int window = 2; window <= list.size(); window *= 2) {
            c_bitonic_sort_spiral(list, window);
            for (int difference = window / 2; difference >= 2; difference /= 2) {
                c_bitonic_sort_difference(list, difference);
            }
        }
    }

    private void c_bitonic_sort_spiral(LinkedList<Integer> list, int size) throws IOException, InternalException {
        for (int i = 0; i < list.size(); i += size) {
            for (int i1 = i, i2 = i + size - 1; i1 < i2; i1++, i2--) {
                c_bitonic_comparator(list, i1, i2);
            }
        }
    }

    private void c_bitonic_sort_difference(LinkedList<Integer> list, int size) throws IOException, InternalException {
        int half = size / 2;
        for (int i = 0; i < list.size(); i += size) {
            for (int j = 0; j < half; j++) {
                c_bitonic_comparator(list, i + j, i + j + half);
            }
        }
    }

    private void c_bitonic_comparator(LinkedList<Integer> result, int i1, int i2) throws IOException, InternalException {
        if (i1 >= i2) {
            throw new InternalException("D_Ausgabe: c_bitonic_comparator: " + i1 + "," + i2 + " <-- ERROR!!!");
        }
        int a = result.get(i1);
        int b = result.get(i2);
        result.set(i1, c_new_var_OR(a, b));
        result.set(i2, c_new_var_AND(a, b));
    }

    // ############################# CLASSES #############################
    private class D_Number_SAT {

        private final int[] unary;

        public D_Number_SAT() {
            this.unary = c_new_var_array(5);
        }

        private D_Number_SAT(LinkedList<Integer> list) {
            this.unary = new int[list.size()];
            for (int i = 0; i < unary.length; i++) {
                this.unary[i] = list.pollFirst();
            }
        }

        public void cnf_post() throws IOException, InternalException {
            for (int i = 1; i < unary.length; i++) {
                // NOT 0 1
                int x = unary[i - 1];
                int y = unary[i];
                c_2(x, -y);
            }
        }

        private int getNr() {
            for (int i = 0; i < unary.length; i++) {
                if (!c_get(unary[i])) {
                    return i;
                }
            }
            return unary.length;
        }

        private D_Number_SAT add(D_Number_SAT that) throws IOException, InternalException {
            LinkedList<Integer> list = Static.convert(this.unary);
            list.addAll(Static.convert(that.unary));
            c_bitonic_sort(list);
            return new D_Number_SAT(list);
        }

        private void cnf_ist_c_stuendig(String c) throws IOException, InternalException {
            int value = Integer.parseInt(c);
            for (int i = 0; i < unary.length; i++) {
                if (i < value) {
                    c_1(+unary[i]);
                } else {
                    c_1(-unary[i]);
                }
            }
        }

        private void cnf_pre() {
            // nichts
        }

        private void cnf_niemals_c_stuendig(String c) throws IOException, InternalException {
            int std = Integer.parseInt(c);
            // not "0"
            if (std == 0) {
                c_1(+unary[0]);
                return;
            }
            // not "1"
            if (std == unary.length) {
                c_1(-unary[std - 1]);
                return;
            }
            // not "10"
            int x = unary[std - 1];
            int y = unary[std];
            c_2(-x, y);
        }

        private void cnf_zwischen_b_und_c_wochenstunden(String b, String c) throws IOException, InternalException {
            int valB = Integer.parseInt(b);
            int valC = Integer.parseInt(c);
            for (int i = 0; i <= unary.length; i++) {
                if (i < valB) {
                    cnf_niemals_c_stuendig(i + "");
                }
                if (i > valC) {
                    cnf_niemals_c_stuendig(i + "");
                }
            }
        }

        private void cnf_gleich(D_Number_SAT b) throws IOException, InternalException {
            int max = Math.max(unary.length, b.unary.length);
            for (int i = 0; i < max; i++) {
                if ((i < unary.length) && (i < b.unary.length)) {
                    c_equal(unary[i], b.unary[i]);
                } else {
                    if (i < unary.length) {
                        c_1(-unary[i]);
                    } else {
                        c_1(-b.unary[i]);
                    }
                }
            }
        }

    }

    private class D_Fach_SAT implements Comparable<D_Fach_SAT> {

        final String kuerzel;
        final String name;
        final D_Number_SAT[] stufe;

        @Override
        public String toString() {
            return name;
        }

        public D_Fach_SAT(String kuerzel, String name) {
            this.kuerzel = kuerzel;
            this.name = name;
            this.stufe = new D_Number_SAT[6];

            // Stunden pro Halbjahr
            for (int s = 0; s < stufe.length; s++) {
                stufe[s] = new D_Number_SAT();
            }

        }

        private void cnf_pre() {
            // Stunden pro Halbjahr
            for (int s = 0; s < stufe.length; s++) {
                stufe[s].cnf_pre();
            }
        }

        private void cnf_post() throws IOException, InternalException {
            // Stunden pro Halbjahr
            for (int s = 0; s < stufe.length; s++) {
                stufe[s].cnf_post();
            }
        }

        @Override
        public int compareTo(D_Fach_SAT b) {
            return Static.compare(kuerzel, b.kuerzel);
        }

        private void cnf_in_der_sek_b_niemals_c_stuendig(String b, String c) throws IOException, InternalException {
            for (int i = 0; i < stufe.length; i++) {
                boolean b1 = b.equals(Static.WILDCARD);
                boolean b2 = b.equals("1") && (i <= 1);
                boolean b3 = b.equals("2") && (i > 1);
                if (b1 || b2 || b3) {
                    stufe[i].cnf_niemals_c_stuendig(c);
                }
            }
        }

        private void cnf_in_der_sek_b_in_der_summe_c_unterrichtsstunden(String b, String c) throws IOException, InternalException {
            LinkedList<D_Number_SAT> list = new LinkedList<>();

            for (int i = 0; i < stufe.length; i++) {
                boolean b1 = b.equals(Static.WILDCARD);
                boolean b2 = b.equals("1") && (i <= 1);
                boolean b3 = b.equals("2") && (i > 1);
                if (b1 || b2 || b3) {
                    list.addLast(stufe[i]);
                }
            }

            while (list.size() >= 2) {
                D_Number_SAT n1 = list.pollFirst();
                D_Number_SAT n2 = list.pollFirst();
                D_Number_SAT n3 = n1.add(n2);
                list.addLast(n3);
            }
            list.pollFirst().cnf_ist_c_stuendig(c);
        }

        private void cnf_in_der_stufe_b_ist_c_stuendig(String b, String c) throws IOException, InternalException {
            int s = Integer.parseInt(b);
            stufe[s - 5].cnf_ist_c_stuendig(c);
        }

    }

}
