package stundentafel_cnf;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.text.ParseException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.TreeMap;
import java.util.TreeSet;
import javax.swing.DefaultListModel;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

/**
 * @author Benjamin A. Bartsch
 */
public final class Data {

    private HashMap<String, Object> map = new HashMap<>();
    private final TreeMap<Integer, D_Berechnung> map_beB = new TreeMap<>();

    final FaListModel fa_list_model = new FaListModel();
    private D_Fach[] fa_Array = new D_Fach[0];

    final BeBListModel beB_list_model = new BeBListModel();
    private D_Berechnung[] beB_array = new D_Berechnung[0];

    final ErBListModel erB_list_model = new ErBListModel();
    private D_Ergebnis[] erB_array = new D_Ergebnis[0];

    public Data() {
        try {
            misc_reset();
        } catch (InternalException | UserException ex) {
        }
    }

    // ##### MISC #####
    synchronized void misc_reset() throws UserException, InternalException {
        map.clear();
        map_beB.clear(); // TODO auch Berechnungen beenden

        // auto load
        fa_create("D", "Deutsch");
        fa_create("GE", "Geschichte");
        fa_create("EK", "Erdkunde");
        fa_create("WIPO", "Wirtschaft/Politik");
        fa_create("M", "Mathematik");
        fa_create("BI", "Biologie");
        fa_create("CH", "Chemie");
        fa_create("PH", "Physik");
        fa_create("E", "Englisch");
        fa_create("F/L", "Franz./Latein");
        fa_create("KU", "Kunst");
        fa_create("MU", "Musik");
        fa_create("RE", "Religion");
        fa_create("SP", "Sport");
        fa_create("WPL", "Wahlpflicht");
        fa_create("ERG", "Ergänzung");

        String regeln = "";
        regeln += "\n100#Fach<D> hat in der Sek.<1> in der Summe<9> Unterrichtsstunden.";
        regeln += "\n100#Fach<D> hat in der Sek.<2> in der Summe<13> Unterrichtsstunden.";
        regeln += "\n102#Fach<D> in der Sek.<2> niemals <0>-stündig.";
        regeln += "\n102#Fach<D> in der Sek.<2> niemals <1>-stündig.";
        regeln += "\n102#Fach<D> in der Sek.<2> niemals <2>-stündig.";
        regeln += "\n102#Fach<D> in der Sek.<2> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<E> hat in der Sek.<1> in der Summe<9> Unterrichtsstunden.";
        regeln += "\n100#Fach<E> hat in der Sek.<2> in der Summe<13> Unterrichtsstunden.";
        regeln += "\n102#Fach<E> in der Sek.<2> niemals <0>-stündig.";
        regeln += "\n102#Fach<E> in der Sek.<2> niemals <1>-stündig.";
        regeln += "\n102#Fach<E> in der Sek.<2> niemals <2>-stündig.";
        regeln += "\n102#Fach<E> in der Sek.<2> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<M> hat in der Sek.<1> in der Summe<9> Unterrichtsstunden.";
        regeln += "\n100#Fach<M> hat in der Sek.<2> in der Summe<13> Unterrichtsstunden.";
        regeln += "\n102#Fach<M> in der Sek.<2> niemals <0>-stündig.";
        regeln += "\n102#Fach<M> in der Sek.<2> niemals <1>-stündig.";
        regeln += "\n102#Fach<M> in der Sek.<2> niemals <2>-stündig.";
        regeln += "\n102#Fach<M> in der Sek.<2> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<F/L> hat in der Sek.<1> in der Summe<0> Unterrichtsstunden.";
        regeln += "\n100#Fach<F/L> hat in der Sek.<2> in der Summe<15> Unterrichtsstunden.";
        regeln += "\n102#Fach<F/L> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<GE> hat in der Sek.<*> in der Summe<8> Unterrichtsstunden.";
        regeln += "\n102#Fach<GE> in der Sek.<*> niemals <1>-stündig.";
        regeln += "\n102#Fach<GE> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<GE> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<GE> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<EK> hat in der Sek.<*> in der Summe<7> Unterrichtsstunden.";
        regeln += "\n102#Fach<EK> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<EK> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<EK> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<WIPO> hat in der Sek.<*> in der Summe<8> Unterrichtsstunden.";
        regeln += "\n102#Fach<WIPO> in der Sek.<*> niemals <1>-stündig.";
        regeln += "\n102#Fach<WIPO> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<WIPO> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<WIPO> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<BI> hat in der Sek.<*> in der Summe<8> Unterrichtsstunden.";
        regeln += "\n102#Fach<BI> in der Sek.<*> niemals <1>-stündig.";
        regeln += "\n102#Fach<BI> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<BI> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<BI> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<CH> hat in der Sek.<1> in der Summe<0> Unterrichtsstunden.";
        regeln += "\n100#Fach<CH> hat in der Sek.<2> in der Summe<7> Unterrichtsstunden.";
        regeln += "\n102#Fach<CH> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<CH> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<CH> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<PH> hat in der Sek.<*> in der Summe<8> Unterrichtsstunden.";
        regeln += "\n102#Fach<PH> in der Sek.<*> niemals <1>-stündig.";
        regeln += "\n102#Fach<PH> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<PH> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<PH> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<KU> hat in der Sek.<*> in der Summe<9> Unterrichtsstunden.";
        regeln += "\n102#Fach<KU> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<KU> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<KU> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<MU> hat in der Sek.<*> in der Summe<8> Unterrichtsstunden.";
        regeln += "\n102#Fach<MU> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<MU> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<MU> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n102#Fach<RE> in der Sek.<*> niemals <0>-stündig.";
        regeln += "\n102#Fach<RE> in der Sek.<*> niemals <1>-stündig.";
        regeln += "\n102#Fach<RE> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<RE> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<RE> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<SP> hat in der Sek.<*> in der Summe<18> Unterrichtsstunden.";
        regeln += "\n102#Fach<SP> in der Sek.<*> niemals <0>-stündig.";
        regeln += "\n102#Fach<SP> in der Sek.<*> niemals <1>-stündig.";
        regeln += "\n102#Fach<SP> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<WPL> hat in der Sek.<*> in der Summe<6> Unterrichtsstunden.";
        regeln += "\n102#Fach<WPL> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n101#Fach<WPL> in der Stufe<5> ist <0>-stündig.";
        regeln += "\n101#Fach<WPL> in der Stufe<6> ist <0>-stündig.";
        regeln += "\n101#Fach<WPL> in der Stufe<7> ist <0>-stündig.";
        regeln += "\n101#Fach<WPL> in der Stufe<8> ist <0>-stündig.";
        regeln += "\n101#Fach<WPL> in der Stufe<9> ist <3>-stündig.";
        regeln += "\n101#Fach<WPL> in der Stufe<10> ist <3>-stündig.";
        regeln += "\n";

        regeln += "\n100#Fach<ERG> hat in der Sek.<*> in der Summe<4> Unterrichtsstunden.";
        regeln += "\n102#Fach<ERG> in der Sek.<*> niemals <5>-stündig.";
        regeln += "\n102#Fach<ERG> in der Sek.<*> niemals <4>-stündig.";
        regeln += "\n102#Fach<ERG> in der Sek.<*> niemals <3>-stündig.";
        regeln += "\n102#Fach<ERG> in der Sek.<*> niemals <2>-stündig.";
        regeln += "\n";

        regeln += "\n103#Stufe<5> hat zwischen <30> und <30> Wochenstunden.";
        regeln += "\n103#Stufe<6> hat zwischen <30> und <30> Wochenstunden.";
        regeln += "\n103#Stufe<7> hat zwischen <30> und <31> Wochenstunden.";
        regeln += "\n103#Stufe<8> hat zwischen <30> und <31> Wochenstunden.";
        regeln += "\n103#Stufe<9> hat zwischen <30> und <31> Wochenstunden.";
        regeln += "\n103#Stufe<10> hat zwischen <30> und <31> Wochenstunden.";
        regeln += "\n";

        regeln += "\n101#Fach<D> in der Stufe<10> ist <3>-stündig.";
        regeln += "\n101#Fach<E> in der Stufe<10> ist <3>-stündig.";
        regeln += "\n101#Fach<M> in der Stufe<10> ist <3>-stündig.";
        regeln += "\n101#Fach<F/L> in der Stufe<10> ist <3>-stündig.";
        regeln += "\n";

        regeln += "\n101#Fach<GE> in der Stufe<10> ist <2>-stündig.";
        regeln += "\n101#Fach<EK> in der Stufe<10> ist <1>-stündig.";
        regeln += "\n101#Fach<WIPO> in der Stufe<10> ist <2>-stündig.";
        regeln += "\n";

        regeln += "\n101#Fach<BI> in der Stufe<10> ist <2>-stündig.";
        regeln += "\n101#Fach<CH> in der Stufe<10> ist <1>-stündig.";
        regeln += "\n101#Fach<PH> in der Stufe<10> ist <2>-stündig.";
        regeln += "\n";

        regeln += "\n101#Fach<KU> in der Stufe<10> ist <1>-stündig.";
        regeln += "\n101#Fach<MU> in der Stufe<10> ist <1>-stündig.";
        regeln += "\n";

        regeln += "\n104#Stufe<7> gleichbehandlung der Fächer <D>, <E>, <M>.";
        regeln += "\n104#Stufe<8> gleichbehandlung der Fächer <D>, <E>, <M>.";
        regeln += "\n104#Stufe<9> gleichbehandlung der Fächer <D>, <E>, <M>.";
        regeln += "\n";

        co_set_regeln(regeln);

        miscRevalidateData();
    }

    synchronized void misc_save(ObjectOutputStream oos) throws IOException {
        oos.writeObject(map);
    }

    synchronized void misc_open(ObjectInputStream ois) throws IOException, ClassNotFoundException {
        map = (HashMap<String, Object>) ois.readObject();
        miscRevalidateData();
    }

    synchronized private void miscRevalidateData() {
        int i;

        // ##### Fa #####
        HashMap<Integer, Object> faMap = fa_all();
        fa_Array = new D_Fach[faMap.size()];
        i = 0;
        for (int id : new TreeSet<>(faMap.keySet())) {
            fa_Array[i] = new D_Fach(id, fa_single_get_kuerzel(id), fa_single_get_name(id));
            i++;
        }
        Arrays.sort(fa_Array);
        fa_list_model.clear(); // TODO okay?
        for (ListDataListener ldl : fa_list_model.getListDataListeners()) {
            ldl.contentsChanged(new ListDataEvent(fa_list_model, ListDataEvent.CONTENTS_CHANGED, 0, fa_list_model.size() - 1));
        }

        // ##### BeB #####
        beB_array = new D_Berechnung[map_beB.size()];
        i = 0;
        for (int beID : map_beB.keySet()) {
            beB_array[i] = map_beB.get(beID);
            i++;
        }
        beB_list_model.clear();
        for (ListDataListener ldl : beB_list_model.getListDataListeners()) {
            ldl.contentsChanged(new ListDataEvent(beB_list_model, ListDataEvent.CONTENTS_CHANGED, 0, beB_list_model.size() - 1));
        }

        // ##### ErB #####
        erB_array = new D_Ergebnis[erB_all().size()];
        i = 0;
        for (int erID : erB_all().keySet()) {
            String data = erB_single_get_data(erID);
            erB_array[i] = new D_Ergebnis(erID, data);
            i++;
        }
        Arrays.sort(erB_array);
        erB_list_model.clear();
        for (ListDataListener ldl : erB_list_model.getListDataListeners()) {
            ldl.contentsChanged(new ListDataEvent(erB_list_model, ListDataEvent.CONTENTS_CHANGED, 0, erB_list_model.size() - 1));
        }

    }

    // ##### FA #####
    synchronized HashMap<Integer, Object> fa_all() {
        if (!map.containsKey("FÄCHER")) {
            map.put("FÄCHER", new HashMap<>());
        }
        return (HashMap<Integer, Object>) map.get("FÄCHER");
    }

    synchronized HashMap<String, Object> fa_single(int id) {
        return (HashMap<String, Object>) fa_all().get(id);
    }

    synchronized String fa_single_get_kuerzel(int id) {
        return (String) fa_single(id).get("KÜRZEL");
    }

    synchronized void fa_single_set_kuerzel(int id, String newKuerzel) {
        fa_single(id).put("KÜRZEL", newKuerzel);
        miscRevalidateData();
    }

    synchronized String fa_single_get_name(int id) {
        return (String) fa_single(id).get("NAME");
    }

    synchronized void fa_single_set_name(int id, String newName) {
        fa_single(id).put("NAME", newName);
        miscRevalidateData();
    }

    synchronized boolean fa_exists(String kuerzel) {
        for (int id : fa_all().keySet()) {
            if (fa_single_get_kuerzel(id).equals(kuerzel)) {
                return true;
            }
        }
        return false;
    }

    synchronized D_Fach fa_create(String newKuerzel) throws UserException, InternalException {
        if (newKuerzel.trim().isEmpty()) {
            throw new UserException("Ungültiger Name für ein Fach!");
        }

        if (fa_exists(newKuerzel)) {
            throw new UserException("Das Fach '" + newKuerzel + "' gibt es bereits!");
        }

        HashMap<Integer, Object> faMap = fa_all();

        int id = 0;
        for (int idOld : faMap.keySet()) {
            id = Math.max(id, idOld);
        }
        id++;

        faMap.put(id, new HashMap<>());
        fa_single_set_kuerzel(id, newKuerzel);
        fa_single_set_name(id, "");

        miscRevalidateData();

        return fa_get_from_array(id);
    }

    synchronized D_Fach fa_create(String newKuerzel, String newName) throws UserException, InternalException {
        if (newKuerzel.trim().isEmpty()) {
            throw new UserException("Ungültiger Name für ein Fach!");
        }

        if (fa_exists(newKuerzel)) {
            throw new UserException("Das Fach '" + newKuerzel + "' gibt es bereits!");
        }

        HashMap<Integer, Object> faMap = fa_all();

        int id = 0;
        for (int idOld : faMap.keySet()) {
            id = Math.max(id, idOld);
        }
        id++;

        faMap.put(id, new HashMap<>());
        fa_single_set_kuerzel(id, newKuerzel);
        fa_single_set_name(id, newName);

        miscRevalidateData();

        return fa_get_from_array(id);
    }

    synchronized D_Fach fa_create_or_get(String faKuerzel) throws UserException, InternalException {
        for (D_Fach fa : fa_Array) {
            if (fa.kuerzel.equals(faKuerzel)) {
                return fa;
            }
        }

        return fa_create(faKuerzel);
    }

    synchronized D_Fach fa_get(String faKuerzel) throws UserException, InternalException {
        for (D_Fach fa : fa_Array) {
            if (fa.kuerzel.equals(faKuerzel)) {
                return fa;
            }
        }
        throw new UserException("Das Fach '" + faKuerzel + "' gibt es nicht!");
    }

    synchronized D_Fach fa_get_from_array(int id) throws InternalException {
        for (int i = 0; i < fa_Array.length; i++) {
            if (fa_Array[i].id == id) {
                return fa_Array[i];
            }
        }
        throw new InternalException("Das Fach mit ID '" + id + "' konnte in 'faArray' nicht gefunden werden!");
    }

    synchronized void fa_del(D_Fach fa) {
        HashMap<Integer, Object> faMap = fa_all();
        faMap.remove(fa.id);

        miscRevalidateData();
    }

    synchronized void fa_del(int faID) {
        HashMap<Integer, Object> faMap = fa_all();
        faMap.remove(faID);

        miscRevalidateData();
    }

    synchronized boolean fa_valid_change_kuerzel(int id, String newKurzel) {
        if (newKurzel.trim().isEmpty()) {
            return false;
        }

        HashMap<Integer, Object> faMap = fa_all();
        for (int idOld : faMap.keySet()) {
            if (idOld != id) {
                if (fa_single_get_kuerzel(id).equals(newKurzel)) {
                    return false;
                }
            }
        }

        return true;
    }

    synchronized TreeSet< String> fa_get_set() {
        TreeSet<String> list = new TreeSet<>();
        for (int faID : fa_all().keySet()) {
            String fach_kuerzel = fa_single_get_kuerzel(faID);
            list.add(fach_kuerzel);
        }
        return list;

    }

    // ##### BE #####
    synchronized D_Berechnung beB_create() throws IOException, UserException, InternalException, ParseException {
        // Berechne maxID von Be und Er
        int id = 0;
        if (!map_beB.isEmpty()) {
            id = map_beB.lastKey();
        }

        for (int erID : erB_all().keySet()) {
            id = Math.max(erID, id);
        }
        id++;

        D_Berechnung be = new D_Berechnung(id, this, null); // fake berechnung --> vars / clauses sind bekannt
        String header = "p cnf " + be.c_vars + " " + be.c_clauses + "\n";
        be = new D_Berechnung(id, this, header); // echte berechnung --> vars / clauses sind bekannt

        map_beB.put(id, be);

        miscRevalidateData();

        return be;
    }

    synchronized void beB_finished(int id) {
        if (!map_beB.containsKey(id)) {
            return; // wurde bereits manuell abgebrochen
        }

        D_Berechnung be = map_beB.get(id);
        String data = be.get_ergebnis();

        map_beB.remove(id);

        erB_create(id, data);
    }

    synchronized void beB_del(int id) {
        D_Berechnung beb = map_beB.remove(id);
        map_beB.remove(beb.beID);
        if (beb.process != null) {
            beb.process.destroyForcibly();
        }
        miscRevalidateData();
    }

    // ##### ER #####
    synchronized HashMap<Integer, Object> erB_all() {
        if (!map.containsKey("ERGEBNISSE_B")) {
            map.put("ERGEBNISSE_B", new HashMap<>());
        }
        return (HashMap<Integer, Object>) map.get("ERGEBNISSE_B");
    }

    synchronized void erB_create(int id, String sData) {
        erB_all().put(id, new HashMap<>());
        erB_single_set_data(id, sData);
        miscRevalidateData();
    }

    synchronized void erB_del(D_Ergebnis er) {
        HashMap<Integer, Object> erMap = erB_all();
        erMap.remove(er.id);
        miscRevalidateData();
    }

    synchronized HashMap<String, Object> erB_single(int id) {
        return (HashMap<String, Object>) erB_all().get(id);
    }

    synchronized void erB_single_set_data(int id, String data) {
        erB_single(id).put("DATA", data);
        miscRevalidateData();
    }

    synchronized String erB_single_get_data(int id) {
        if (!erB_single(id).containsKey("DATA")) {
            erB_single(id).put("DATA", "KEINE DATEN VORHANDEN");
        }
        return (String) erB_single(id).get("DATA");
    }

    // ##### CO #####
    synchronized void co_set_regeln(String value) {
        map.put("REGELN", value);
    }

    synchronized String co_get_regeln() {
        if (!map.containsKey("REGELN")) {
            map.put("REGELN", "");
        }
        return (String) map.get("REGELN");
    }

    synchronized LinkedList<String> co_get_regel_liste() {
        LinkedList<String> list = new LinkedList<>();
        for (String line : co_get_regeln().split("[\\r\\n]+")) {
            String[] split2 = line.split("\\#");
            if (split2.length == 2) {
                try {
                    int nr = Integer.parseInt(split2[0]);
                    list.addLast(line);
                } catch (NumberFormatException ex) {
                }
            }
        }
        return list;
    }

    /*
    synchronized TreeSet< String> fa_get_set() {
        TreeSet<String> list = new TreeSet<>();
        for (int kuID : ku_all().keySet()) {
            String fach_kuerzel = ku_single_get_fach(kuID);
            list.add(fach_kuerzel);
        }
        return list;

    }
     */
    // ##### class #####
    class FaListModel extends DefaultListModel<D_Fach> {

        @Override
        public int getSize() {
            return fa_Array.length;
        }

        @Override
        public D_Fach getElementAt(int index) {
            return fa_Array[index];
        }

    }

    class BeBListModel extends DefaultListModel<D_Berechnung> {

        @Override
        public int getSize() {
            return beB_array.length;
        }

        @Override
        public D_Berechnung getElementAt(int index) {
            return beB_array[index];
        }

    }

    class ErBListModel extends DefaultListModel<D_Ergebnis> {

        @Override
        public int getSize() {
            return erB_array.length;
        }

        @Override
        public D_Ergebnis getElementAt(int index) {
            return erB_array[index];
        }
    }

}
