package stundentafel_cnf;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.ItemListener;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.prefs.Preferences;
import javax.swing.JComboBox;

/**
 * @author Benjamin A. Bartsch
 */
public class Static {

    static final String TITLE = "bar.Stundentafel_CNF (2020_10_02a) (bar@gymnasium-eickel.herne.de)";

    private static final HashMap<String, Integer> MAP = new HashMap<>();

    static {
        MAP.put("D", 1);
        MAP.put("E", 2);
        MAP.put("M", 3);
        MAP.put("F/L", 4);
        
        MAP.put("GE", 5);
        MAP.put("EK", 6);
        MAP.put("WIPO", 7);

        MAP.put("BI", 8);
        MAP.put("CH", 9);
        MAP.put("PH", 10);

        MAP.put("KU", 11);
        MAP.put("MU", 12);

        MAP.put("RE", 13);
        MAP.put("SP", 14);
        MAP.put("WPL", 15);

        MAP.put("ERG", 99);
    }


    static final Charset CHAR_SET_ISO = StandardCharsets.ISO_8859_1;
    static final Charset CHAR_SET_UTF_8 = StandardCharsets.UTF_8;
    static final SimpleDateFormat SDF = new SimpleDateFormat("dd.MM.yyyy");

    static final String[] WOCHENTAG = new String[]{"", "Mo", "Di", "Mi", "Do", "Fr", "Sa", "So"};

    static final int PLAN_ROWS = 9;
    static final int PLAN_ROWS2 = PLAN_ROWS + 2;
    static final int PLAN_COLS = 5;
    static final int JAHRGANG_MIN = 5;
    static final int JAHRGANG_MAX = 9;

    static final Font FONT_TABS = new Font(Font.SANS_SERIF, Font.BOLD, 20);
    static final Font FONT_AUSGABE = new Font("Monospaced", Font.BOLD, 12);

    static Font FONT_DEFAULT = new Font("", Font.PLAIN, 12);
    static Font FONT_DEFAULT_SMALLER = new Font("", Font.PLAIN, 9);
    static Font FONT_TITLED_BORDER = new Font("", Font.PLAIN, 9);
    static Font FONT_TEXT_AREA = new Font("", Font.PLAIN, 12);
    static Font FONT_CHECK_BOX = new Font("", Font.PLAIN, 12);

    static final Color COLOR_EDITABLE = Color.WHITE;
    static final Color COLOR_ERROR = new Color(200, 100, 0);
    static final Color COLOR_NOT_EDITABLE = Color.LIGHT_GRAY;
    static final String WILDCARD = "*";

    static String getStandardPfadSpeichern() {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        return pref.get("StandardSpeicherpfad", "");
    }

    static void setStandardPfadSpeichern(String value) {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        pref.put("StandardSpeicherpfad", value);
    }

    static String getStandardPfadKurs42() {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        return pref.get("StandardSpeicherpfadKurs42", "");
    }

    static void setStandardPfadKurs42(String value) {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        pref.put("StandardSpeicherpfadKurs42", value);
    }

    static String get_path_crypto() {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        return pref.get("StandardPfadCrypto", "");
    }

    static void set_path_crypto(String pfad) {
        Preferences pref = Preferences.userNodeForPackage(Static.class);
        pref.put("StandardPfadCrypto", pfad);
    }

    static String get_stretched(String s, int length) {
        while (s.length() < length) {
            s = s + " ";
        }
        return s;
    }

    static void remove_all_listener(JComboBox cb) {
        for (ItemListener il : cb.getItemListeners()) {
            cb.removeItemListener(il);
        }
    }

    static String extract(String text, int stelle) {

        int index1 = -1;
        for (int i = 0; i < stelle; i++) {
            index1 = text.indexOf('<', index1 + 1);
        }
        int index2 = text.indexOf('>', index1 + 1);

        String result = text.substring(index1 + 1, index2);
        return result;
    }

    static String replace(String text, int stelle, String inhalt) throws UserException, InternalException {
        int index1 = -1;
        for (int i = 0; i < stelle; i++) {
            index1 = text.indexOf('<', index1 + 1);
        }
        int index2 = text.indexOf('>', index1 + 1);
        int length = text.length();

        if (index1 < 0) {
            throw new InternalException("Stelle " + stelle + " in " + text + " nicht auffindbar!");
        }

        return text.substring(0, index1) + "<" + inhalt + ">" + text.substring(index2 + 1, length);
    }

    static String get_wochentag(int c) {
        return WOCHENTAG[c];
    }

    static String datum_naechster_tag(String datum) throws ParseException {
        Calendar c = Calendar.getInstance();
        Date date = SDF.parse(datum);
        c.setTime(date);
        c.add(Calendar.DATE, 1);
        Date date2 = c.getTime();
        return SDF.format(date2);
    }

    static String datum_vorheriger_tag(String datum) throws ParseException {
        Calendar c = Calendar.getInstance();
        Date date = SDF.parse(datum);
        c.setTime(date);
        c.add(Calendar.DATE, -1);
        Date date2 = c.getTime();
        return SDF.format(date2);
    }

    static String datum_formatiere(String datum) throws ParseException {
        Date date = SDF.parse(datum);
        return SDF.format(date);
    }

    static String datum_get_montag_in_woche(String datum) throws ParseException {
        Calendar c = Calendar.getInstance();
        Date date = SDF.parse(datum);
        c.setTime(date);
        while (c.get(Calendar.DAY_OF_WEEK) != Calendar.MONDAY) {
            c.add(Calendar.DATE, -1);
        }
        return SDF.format(c.getTime());
    }

    static String datum_get_sonntag_in_woche(String datum) throws ParseException {
        Calendar c = Calendar.getInstance();
        Date date = SDF.parse(datum);
        c.setTime(date);
        while (c.get(Calendar.DAY_OF_WEEK) != Calendar.SUNDAY) {
            c.add(Calendar.DATE, +1);
        }
        return SDF.format(c.getTime());
    }

    static int datum_compare(String datumVon, String datumBis) throws ParseException, UserException {
        String d1 = datum_formatiere(datumVon);
        String d2 = d1;
        datumBis = datum_formatiere(datumBis);

        if (d1.equals(datumBis)) {
            return 0;
        }

        for (int i = 0; i < 1000; i++) {
            d1 = datum_vorheriger_tag(d1);
            d2 = datum_naechster_tag(d2);
            if (d1.equals(datumBis)) {
                return +1;
            }
            if (d2.equals(datumBis)) {
                return -1;
            }
        }

        throw new UserException("Datumsdifferenz zu groß!");
    }

    static String datum_get_wochentag(String datum) throws ParseException {
        Date date = SDF.parse(datum);
        return new SimpleDateFormat("EEEE").format(date);
    }

    static int datum_get_plan_nr(String sDatum, boolean drehen) throws ParseException {
        Date date = SDF.parse(sDatum);
        String sKW = new SimpleDateFormat("w").format(date);
        int woche_ab = Integer.parseInt(sKW) % 2;
        if (drehen) {
            woche_ab = 1 - woche_ab;
        }
        return woche_ab;
    }

    static Rectangle get_centered_rectangle() {
        Dimension dim1 = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension dim2 = new Dimension((int) (dim1.width * 0.75), (int) (dim1.height * 0.75));
        int x = (int) ((dim1.getWidth() - dim2.width) / 2);
        int y = (int) ((dim1.getHeight() - dim2.height) / 2);
        return new Rectangle(x, y, dim2.width, dim2.height);
    }

    static LinkedList<String> htmlEscape(LinkedList<String> a) {
        LinkedList<String> b = new LinkedList<>();
        for (String s : a) {
            s = s.replaceAll("Ä", "&Auml;");
            s = s.replaceAll("ä", "&auml;");
            s = s.replaceAll("Ö", "&Ouml;");
            s = s.replaceAll("ö", "&ouml;");
            s = s.replaceAll("Ü", "&Uuml;");
            s = s.replaceAll("ü", "&uuml;");
            s = s.replaceAll("ß", "&szlig;");
            b.addLast(s);
        }
        return b;
    }

    static void toHtmlTable(String[][] planS, String[] header, LinkedList<String> html) {
        int w = 100 / (header.length + 1);

        int ROWS = planS.length;
        int COLS = planS[0].length;
        String styleTable = "style=\"width: 100%; table-layout:fixed; border: 1px solid black; border-collapse: collapse;\"";
        String sStyle1 = "style=\"width:" + (2 * w) + "%; border: 1px solid black;border-collapse: collapse;\"";
        String sStyle2 = "style=\"width:" + (w) + "%; border: 1px solid black;border-collapse: collapse;\" colspan=\"2\"";

        String sTD1 = "<td " + sStyle1 + ">";
        String sTD2 = "<td " + sStyle2 + ">";

        String sTH1 = "<th " + sStyle1 + ">";
        String sTH2 = "<th " + sStyle2 + ">";
        String sTR = "<tr>";

        html.addLast("<table " + styleTable + ">");

        // table header
        html.addLast(sTR);
        String sHeader = "";
        for (int i = 0; i < header.length; i++) {
            sHeader = sHeader + (i == 0 ? sTH1 + header[i] : sTH2 + header[i]);
        }
        html.addLast(sHeader);
        html.addLast("</tr>");

        // rows
        for (int r = 0; r < ROWS; r++) {
            html.addLast(sTR);
            html.addLast(sTD1 + planS[r][0] + "</td>");
            for (int c = 1; c < COLS; c++) {
                html.addLast(sTD2 + planS[r][c] + "</td>");
            }
            html.addLast("</tr>");
        }

        html.addLast("</table>");
    }

    static void toHtmlTable(String[][][] planS, LinkedList<String> html, String[] header) {
        int ROWS = planS.length;
        int COLS = planS[0].length;
        String styleTable = "style=\"width: 100%; table-layout:fixed; border: 1px solid black; border-collapse: collapse;\"";
        String sStyle1 = "style=\"width:9%; border: 1px solid black;border-collapse: collapse;\"";
        String sStyle2 = "style=\"width:18%; border: 1px solid black;border-collapse: collapse;\" colspan=\"2\"";

        String sTD1 = "<td " + sStyle1 + ">";
        String sTD2 = "<td " + sStyle2 + ">";

        String sTH1 = "<th " + sStyle1 + ">";
        String sTH2 = "<th " + sStyle2 + ">";
        String sTR = "<tr>";

        html.addLast("<table " + styleTable + ">");

        // table header
        html.addLast(sTR);
        String sHeader = "";
        for (int i = 0; i < header.length; i++) {
            sHeader = sHeader + (i == 0 ? sTH1 + header[i] : sTH2 + header[i]);
        }
        html.addLast(sHeader);
        html.addLast("</tr>");

        // rows
        for (int r = 0; r < ROWS; r++) {
            html.addLast(sTR);
            html.addLast(sTD1 + (r + 1) + ". </td>");
            for (int c = 0; c < COLS; c++) {
                if (planS[r][c][0].equals(planS[r][c][1])) {
                    html.addLast(sTD2 + planS[r][c][0] + "</td>");
                } else {
                    if (planS[r][c][0].trim().isEmpty()) {
                        planS[r][c][0] = "-----";
                    }
                    if (planS[r][c][1].trim().isEmpty()) {
                        planS[r][c][1] = "-----";
                    }
                    html.addLast(sTD1 + planS[r][c][0] + "</td>");
                    html.addLast(sTD1 + planS[r][c][1] + "</td>");
                }
            }
            html.addLast("</tr>");
        }

        html.addLast("</table>");
    }

    static String htmlEscapeSpitzeKlammern(String line) {
        line = line.replaceAll("<", "&lt;");
        line = line.replaceAll(">", "&gt;");
        return line;
    }

    static String htmlEscapeZeilenumbruchZuBr(String line) {
        line = line.replace("\n", "<br>").replace("\r", "<br>");
        return line;
    }

    static LinkedList<Integer> convert(int[] a) {
        LinkedList<Integer> list = new LinkedList<>();
        for (int i = 0; i < a.length; i++) {
            list.addLast(a[i]);
        }
        return list;
    }

    static int compare(String kuerzel1, String kuerzel2) {
        int val1 = 999;
        int val2 = 999;
        if (MAP.containsKey(kuerzel1)) {
            val1 = MAP.get(kuerzel1);
        }
        if (MAP.containsKey(kuerzel2)) {
            val2 = MAP.get(kuerzel2);
        }
        if (val1 < val2) {
            return -1;
        }
        if (val1 > val2) {
            return +1;
        }
        int cmp = kuerzel1.compareTo(kuerzel2);
        return cmp;
    }

}
